#include "PovDot.h"
#include <Application.h>
#include <Bitmap.h>
#include <TranslationUtils.h>
#include <stdlib.h>
#include <string.h>
#define PI 3.14159265

void display_usage() {
	printf("PovDot for BeOS (1/22/00) by Daniel Switkin software@switkin.com\n");
	printf("Home page: http://www.switkin.com/software/utilities\n\n");
	printf("Generate a POV-Ray scene of spheres in a 2D array based on\n");
	printf("the color and position of pixels in an image.\n");
	printf("\nUsage: povdot image scene[.pov]\n");
	printf("       povdot image scene[.pov] -scount [-v] [-zvalue]\n");
	printf("       povdot image scene[.pov] -himage [-zvalue]\n\n");
	printf("    -s create this many sine waves through the array\n");
	printf("    -v make the sine wave vertical (default horizontal)\n");
	printf("    -h use a second image as a height field for the spheres\n");
	printf("    -z the scale of the wall of spheres in the z axis (default is 1)\n");
	exit(1);
}

int main(int argc, char **argv) {
	if (argc < 3) display_usage();
	
	char imagename[255];
	char scenename[255];
	char heightname[255];
	strcpy(imagename, argv[1]);
	strcpy(scenename, argv[2]);
	if (strstr(scenename, ".pov") == NULL) {
		strcat(scenename, ".pov");
	}
	
	BApplication app("application/x-vnd.Jules-PovDot");
	PovDot *povdot = new PovDot(imagename, scenename);
	if (povdot->fatalerror) {
		delete povdot;
		return 1;
	}
	
	if (argc == 3) povdot->NormalSpheres();
	else {
	    if (strstr(argv[3], "-s") != NULL) {
            float count = atof(argv[3] + 2);
            bool vertical = false;
            float zsize = 1.0;
            for (int x = 4; x < argc; x++) {
            	if (strstr(argv[x], "-v") != NULL) vertical = true;
            	else if (strstr(argv[x], "-z") != NULL) zsize = atof(argv[x] + 2);
            }
	        povdot->UseSine(count, vertical, zsize);
        } else if (strstr(argv[3], "-h") != NULL) {
            strcpy(heightname, argv[3] + 2);
            if (argc == 5) {
                if (strstr(argv[4], "-z") != NULL) {
                    float height = atof(argv[4] + 2);
                    if (povdot->UseHeightMap(heightname, height) == 1) remove(scenename);
                } else {
                    printf("Illegal option, so exiting\n");
                    delete povdot;
                    exit(1);
                }
            } else if (povdot->UseHeightMap(heightname, 1.0) == 1) remove(scenename);
        } else {
            printf("Illegal options, so exiting\n");
        }
    }
	delete povdot;
	return 0;
}

PovDot::PovDot(char *imagename, char *scenename) {
	fatalerror = false;
	
	bitmap = BTranslationUtils::GetBitmapFile(imagename);
	if (bitmap == NULL) {
		printf("Could not open or translate %s\n", imagename);
		fatalerror = true;
		return;
	}
	if (bitmap->ColorSpace() != B_RGB32 && bitmap->ColorSpace() != B_RGBA32) {
		printf("Unusable color space returned\n");
		fatalerror = true;
		return;
	}
	width = bitmap->Bounds().IntegerWidth() + 1;
	height = bitmap->Bounds().IntegerHeight() + 1;
	
	if ((fp = fopen(scenename, "w+")) == NULL) {
		printf("Failed to open %s\n", scenename);
		fatalerror = true;
		return;
	}
	
	fprintf(fp, "#include \"colors.inc\"    // Standard Color definitions\n");
	fprintf(fp, "#include \"textures.inc\"    // Standard Texture definitions\n\n");
	fprintf(fp, "camera {\n");
	if (width > height) fprintf(fp, "\tlocation <0, 0, %d>\n", (width + 5) * -1);
    else fprintf(fp, "\tlocation <0, 0, %d>\n", (height + 5) * -1);
	fprintf(fp, "\tlook_at <0, 0, 0>\n}\n\n");
	fprintf(fp, "light_source {\n");
	fprintf(fp, "\t0*x\n");
	fprintf(fp, "\tcolor red 1.0 green 1.0 blue 1.0\n");
	fprintf(fp, "\ttranslate <0, 0, -50>\n}\n\n");
	fprintf(fp, "// Define objects and textures\n");
	fprintf(fp, "#declare tgadot=sphere { <0, 0, 0>, 0.4 }\n");
	fprintf(fp, "#declare tgatex=texture { finish { ambient 0.6 phong 0.3 } }\n\n");
    fprintf(fp, "// Now declare the object\n");
    fprintf(fp, "union {\n");
}

PovDot::~PovDot() {
	delete bitmap;
    fclose(fp);
}

int PovDot::NormalSpheres() {
	int ypos = height / 2;
	uchar *bits = (uchar *)bitmap->Bits();
	if (bits == NULL) {
		printf("Bitmap data was NULL\n");
		return 1;
	}
	
	for (int y = 0; y < height; y++) {
		int xpos = (width / 2) * -1;
		for (int x = 0; x < width; x++) {
			float b = bits[x * 4] / 255.0;
			float g = bits[x * 4 + 1] / 255.0;
			float r = bits[x * 4 + 2] / 255.0;
			
			fprintf(fp, "\tobject {\n");
            fprintf(fp, "\t\ttgadot translate <%d, %d, 0>\n", xpos, ypos);
            fprintf(fp, "\t\ttexture {\n");
            fprintf(fp, "\t\t\ttgatex pigment { color rgb <%5.4f, %5.4f, %5.4f> }\n", r, g, b);
            fprintf(fp, "\t\t}\n");
            fprintf(fp, "\t}\n");
            xpos++;
		}
		ypos--;
		bits += bitmap->BytesPerRow();
	}
	fprintf(fp, "}\n\n");
	return 0;
}

int PovDot::UseSine(float count, bool vertical, float zsize) {
	float zpos;
	int ypos = height / 2;
	uchar *bits = (uchar *)bitmap->Bits();
	if (bits == NULL) {
		printf("Bitmap data was NULL\n");
		return 1;
	}
	
	for (int y = 0; y < height; y++) {
		int xpos = (width / 2) * -1;
		if (vertical) {
			zpos = ((float)y / height) * count * 2 * PI;
			zpos = sin(zpos) * zsize;
		}
		for (int x = 0; x < width; x++) {
			float b = bits[x * 4] / 255.0;
			float g = bits[x * 4 + 1] / 255.0;
			float r = bits[x * 4 + 2] / 255.0;
			
			fprintf(fp, "\tobject {\n");
			if (!vertical) {
				zpos = ((float)x / width) * count * 2 * PI;
			    zpos = sin(zpos) * zsize;
			}
		    fprintf(fp, "\t\ttgadot translate <%d, %d, %5.4f>\n", xpos, ypos, zpos);
            fprintf(fp, "\t\ttexture {\n");
            fprintf(fp, "\t\t\ttgatex pigment { color rgb <%5.4f, %5.4f, %5.4f> }\n", r, g, b);
            fprintf(fp, "\t\t}\n");
            fprintf(fp, "\t}\n");
            xpos++;
		}
		ypos--;
		bits += bitmap->BytesPerRow();
	}
	fprintf(fp, "}\n\n");
	return 0;
}

int PovDot::UseHeightMap(char *name, float zsize) {
	uchar *bits = (uchar *)bitmap->Bits();
	if (bits == NULL) {
		printf("Bitmap data was NULL\n");
		return 1;
	}
    
    BBitmap *heightfield = BTranslationUtils::GetBitmapFile(name);
    if (heightfield == NULL) {
    	printf("Could not open or translate %s\n", name);
    	return 1;
    }
    
    int heightfield_width = heightfield->Bounds().IntegerWidth() + 1;
    int heightfield_height = heightfield->Bounds().IntegerHeight() + 1;
    uchar *heightbits = (uchar *)heightfield->Bits();
    if (heightbits == NULL || (heightfield->ColorSpace() != B_RGB32 &&
    	heightfield->ColorSpace() != B_RGBA32)) {
    	
    	printf("Out of memory or bad color space\n");
    	delete heightfield;
    	return 1;
    }
    
    if (heightfield_width < width || heightfield_height < height) {
    	printf("Dimensions of height field are too small\n");
    	delete heightfield;
    	return 1;
    }

    int ypos = height / 2;
	for (int y = 0; y < height; y++) {
		int xpos = (width / 2) * -1;
		for (int x = 0; x < width; x++) {
			float b = bits[x * 4] / 255.0;
			float g = bits[x * 4 + 1] / 255.0;
			float r = bits[x * 4 + 2] / 255.0;
			
			float zpos = heightbits[x * 4] + (heightbits[x * 4 + 1] << 1) + heightbits[x * 4 + 2];
			zpos = zpos / 4.0 / 255.0 * zsize * -1;
			fprintf(fp, "\tobject {\n");
            fprintf(fp, "\t\ttgadot translate <%d, %d, %5.4f>\n", xpos, ypos, zpos);
            fprintf(fp, "\t\ttexture {\n");
            fprintf(fp, "\t\t\ttgatex pigment { color rgb <%5.4f, %5.4f, %5.4f> }\n", r, g, b);
            fprintf(fp, "\t\t}\n");
            fprintf(fp, "\t}\n");
            xpos++;
		}
		ypos--;
		bits += bitmap->BytesPerRow();
		heightbits += heightfield->BytesPerRow();
	}
	fprintf(fp, "}\n\n");
	delete heightfield;
	return 0;
}